/*
 *  This file is provided for use with the unix-socket-faq.  It is public
 *  domain, and may be copied freely.  There is no copyright on it.  The
 *  original work was by Vic Metcalfe (vic@brutus.tlug.org), and any
 *  modifications made to that work were made with the understanding that
 *  the finished work would be in the public domain.
 *
 *  If you have found a bug, please pass it on to me at the above address
 *  acknowledging that there will be no copyright on your work.
 *
 *  The most recent version of this file, and the unix-socket-faq can be
 *  found at http://www.interlog.com/~vic/sock-faq/.
 */

#include "sockhelp.h"
#include "options.h"
#include "list.h"
#include "aClient.h"
#include "extern.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdlib.h>

char *PACKAGE = "GameServ";
char *VERSION = "1.1.2";

int sock;
long timestamp;

List<aClient> clients;

void save_timestamp();
void load_timestamp();

int main(int argc, char *argv[])
{
  char buffer[1024], buf[1024];
  int connected = 1;
  char *cmd, *source = NULL;
  srand(time(NULL));
  
  load_config_file();

  if (argc == 1) {
    argc = 3;
    argv[1] = remoteserver;
    argv[2] = remoteport;
  }
  if (argc != 3) {
    fprintf(stderr,"Usage:  tcpclient host port\n");
    fprintf(stderr,"where host is the machine which is running the\n");
    fprintf(stderr,"tcpserver program, and port is the port it is\n");
    fprintf(stderr,"listening on.\n");
    exit(EXIT_FAILURE);
  }
  ignore_pipe();
  sock = make_connection(argv[2], SOCK_STREAM, argv[1]);
  if (sock == -1) {
    fprintf(stderr,"make_connection failed.\n");
    unload_config_file();
    return -1;
  }

#ifdef UNREAL
	raw("PROTOCTL NICKv2 VHP");
	raw("PASS :%s", remotepass);
	raw("SERVER %s 1 :%s", servername, servername);
	raw("NICK %S 1 %d %S %s %s %d +owghraAxNt %s :%s v%s", time(NULL), gshost, 
		servername, time(NULL), gshost, PACKAGE, VERSION);
	raw(":%S JOIN %s", c_Forest);
	raw(":%S MODE %s +mtn", c_Forest);
#elif defined(BAHAMUT)
	raw("PASS %s :TS", remotepass);
	raw("SERVER %s 1 :%s", servername, servername);
        raw("NICK %S 1 %d +o %s %s %s 0 :GameServ", time(NULL), gsident, gshost, 
		servername);
	raw(":%s SJOIN %d %d %s +mnt :@%S", servername, time(NULL), time(NULL), c_Forest);
#endif
	raw(":%S MODE %s +o %S", c_Forest);
	raw(":%S TOPIC %s :%s", c_Forest, c_ForestTopic);

  sock_gets(sock,buffer,sizeof(buffer)-1); /* -1 added thanks to
    David Duchene <dave@ltd.com> for pointing out the possible
    buffer overflow resulting from the linefeed added below. */


  printf("Server: %s\n",buffer);
  init_monsters();
  init_masters();
  load_gs_dbase();
  load_timestamp();
  int loadtime = time(NULL);
  bool loaded = false;

  while (connected) {
      if (sock_gets(sock,buffer,sizeof(buffer)) == -1) {
        connected = 0;
      }
	strcpy(buf, buffer);

        if (buffer[0] == ':')
	{
	    source = strtok(buf, " ");
	    cmd = strtok(NULL, " ");
	}
	else
	    cmd = strtok(buf, " ");

	cout << "Server: " << buffer << endl << flush;

	// Wait five seconds then we're loaded.
        if (!loaded)
	{
	    if (time(NULL) >= 5 + loadtime)
		loaded = true;
	}

	if (stricmp(cmd, "PING") == 0) {
	    char *timestamp;
	    timestamp = strtok(NULL, "");
	    raw("PONG %s", timestamp);
	} else if (stricmp(cmd, "VERSION") == 0) {
	    char *server;
	    server = strtok(NULL, " ");
	    server++;
//351 GameServ Unreal3.2-beta18. irc.the-irc.org :FhiXOo [Linux shell.the-irc.org 
//2.4.16-010stab017.17.777-smp #1 SMP Wed Mar 19 16:25:17 MSK 2003 i686 unknown=2303]

	    raw(":%s 351 %s  %s %s. %s", servername, source+1, PACKAGE, VERSION, servername);
	} else if (strncmp(cmd, "NICK", 4) == 0) {
	    if (buffer[0] == ':')
	    {
		aClient *tempPtr;
		if ((tempPtr = find((source + 1))))
		{
		    char *nick;
		    nick = strtok(NULL, " ");
		    tempPtr->setNick(nick);
		}
	    }
	    else
	    {
		char *nick;
		aClient *newuser;
		nick = strtok(NULL, " ");
		newuser = new aClient(nick);
		if (loaded)
		    notice(s_GameServ, nick, "Hello, %s! This network utilizes a services package called GameServ. For info on how to play the game, type /msg %S help.");

		clients.insertAtBack(newuser);
		delete newuser;
	    }
	} else if (stricmp(cmd, "QUIT") == 0) {
	    aClient *quitter;
	    if ((quitter = find(source + 1)))
		clients.remove(quitter);
	    if ((quitter = findplayer(source + 1)))
		quitter->setNick("NULL");

	} else if (stricmp(cmd, "PRIVMSG") == 0) {
	    char *rest, *dest;
	    dest = strtok(NULL, " ");
	    rest = strtok(NULL, "");
	    if (strnicmp(dest, s_GameServ, strlen(s_GameServ)) == 0)
		gameserv(source, rest);
	    else if (stricmp(dest, c_Forest) == 0)
		forest(source, rest);
	} else if (stricmp(cmd, "JOIN") == 0) {
	    char *channel;
	    channel = strtok(NULL, " ");
	    if (stricmp(channel, c_Forest) == 0 && is_playing(source + 1))
		raw(":%S MODE %s +v %s", c_Forest, (source + 1));

	#if defined(BAHAMUT)
	} else if (stricmp(cmd, "SJOIN") == 0) {
	    char *channel, *nick;
	    strtok(NULL, " "); // Ignore the TS
	    strtok(NULL, " "); // Ignore the TS
	    channel = strtok(NULL, " ");
	    strtok(NULL, " ");
	    nick = strtok(NULL, " ");
	    nick++; // Get rid of the :
	    if (stricmp(channel, c_Forest) == 0 && is_playing(nick))
		raw(":%S MODE %s +v %s", channel, nick);
	#endif

	} else {
	   // cout << "Unrecognized Message: cmd = " << cmd << setw(30) << "source = " << 
	   //	    source << endl;
	}
  }
  save_gs_dbase();
  save_timestamp();
  delete_monsters();
  delete_masters();

  printf("<CLOSED>\n");
  close(sock);
  unload_config_file();
  return 0;
}

aClient *find(char *nick)
{
	return findbynick(nick);
}

aClient *find(const char *nick)
{
	return findbynick(nick);
}


aClient *findbynick(char *nick)
{
    ListNode <aClient> *newPtr;
    newPtr = clients.First();

    aClient *client = NULL;

    while (newPtr)
    {
	client = newPtr->getData();
	if (stricmp(client->getNick(), nick) == 0)
	    return client;
	client = NULL;
	newPtr = newPtr->Next();
    }
    return client;    
}

aClient *findplayer(const char *name)
{
    ListNode <aClient> *newPtr;
    Player *p = NULL;

    for (newPtr = players.First(); newPtr; newPtr = newPtr->Next())
    {
	p = newPtr->getData()->stats;
	if (stricmp(p->name, name) == 0)
	    return newPtr->getData();
	p = NULL;
    }
    return NULL;
}

aClient *findbynick(const char *nick)
{
    ListNode <aClient> *newPtr;
    newPtr = clients.First();

    aClient *client = NULL;

    while (newPtr)
    {
	client = newPtr->getData();
	if (stricmp(client->getNick(), nick) == 0)
	    return client;
	client = NULL;
	newPtr = newPtr->Next();
    }
    return client;    
}

void load_timestamp()
{
    ifstream infile;

    infile.open(".gstimestamp");

    if (infile.fail())
    {
	cout << "Error opening .gstimestamp" << endl;
	cout << "Generating new timestamp" << endl;
	generate:
	timestamp = midnight();
        save_timestamp();
	return;
    }

    infile >> timestamp;
    infile.close();
    if (timestamp < 1000000)
	goto generate;
}

void save_timestamp()
{
    ofstream outfile;

    outfile.open(".gstimestamp");

    if (outfile.fail())
    {
	cout << "Error creating new file." << endl;
	return;
    }

    outfile << timestamp << endl;

    outfile.close();
}

long int midnight(long int offset)
{
    return (time(NULL) - (time(NULL) % 86400)) + (offset * 3600);
}
